from genomics.burrows import *
from genomics.global_alignment import *
import logging

logger = logging.getLogger(__name__)


def reverse_and_complement(this):
    complements = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}
    return ''.join(map(lambda x: complements[x], this[::-1]))


def get_seed(seed_length, read):
    return read[0: seed_length]


# make coarse then do fine refine
def seed_and_extend(scoring_points, references, reads, margin=3, seed_length=10, reference_index=0):

    # init
    reference = references[reference_index]+'$'

    logger.info('Starting bwt via bwm')
    bwt_reference = bwt_via_bwm(reference)
    logger.info('Finished bwt via bwm')

    occ_matrix, tots = make_occurrences_matrix(bwt_reference)
    c = first_col(tots)
    suff_arr = make_suffix_array(reference)
    result = []
    for read in reads:
        # for each read extract substrins - seeds and reverse complement
        rc_read = reverse_and_complement(read)

        seed = get_seed(seed_length, read)
        rc_seed = get_seed(seed_length, rc_read)

        result.extend(extract(c, margin, occ_matrix, read, reference, seed, seed_length, scoring_points, suff_arr))
        result.extend(extract(c, margin, occ_matrix, rc_read, reference, rc_seed, seed_length, scoring_points, suff_arr))

    return sorted(result, key=lambda tup: tup[2], reverse=True)


def extract(c, margin, occ_matrix, read, reference, seed, seed_length, scoring_points, suff_arr):
    start, end = calculate_start_end_range(c, occ_matrix, seed)
    if (start, end) == (-1, -1):
        return []
    seed_positions = find_all_query_positions_in_word_via_suffix_arr(start, end, suff_arr)
    # ref     "ctagtcgt agctagctgatcg"
    # read    "gctgt cgtc"
    # s       "gctgt"
    result = []
    for start_of_seed in seed_positions:
        # start/end of reference
        start_pos = start_of_seed + seed_length
        end_pos = start_pos + len(read) - seed_length + margin + 1
        end_pos = len(reference) if end_pos > len(reference) else end_pos

        distances, alignment_score = global_alignment(read[seed_length:], reference[start_pos:end_pos],
                                                      scoring_points)
        alignment, transcript = traceback(read[seed_length:], reference[start_pos:end_pos], distances,
                                          scoring_points)
        result.append((start, end, alignment_score, transcript))
    return result



import numpy
import logging

logger = logging.getLogger(__name__)


def scoring_matrix_inplace(this_c, that_c, scoring_points):
    """ Generate a scoring matrix such that matching has the highest value, transitions are penalised,
    transversions are penalised even more, and gaps are penalised the most. Reminder: A-G, C-T transitions,
    other combinations ase transversions. """
    if this_c == that_c:
        return scoring_points['M']
    if this_c == '_' or that_c == '_':
        return scoring_points['G']
    maxb, minb = max(this_c, that_c), min(this_c, that_c)
    if (minb == 'A' and maxb == 'G') or (minb == 'C' and maxb == 'T'):
        return scoring_points['Ti']
    else:
        return scoring_points['Tv']


def global_alignment(this, that, scoring_points):
    """ Firstly, fill in V(i, 0) and V(0, j) with gap values, then the whole distance matrix. """
    distance_matrix = numpy.zeros((len(this) + 1, len(that) + 1), dtype=int)

    for i in range(1, len(this) + 1):
        distance_matrix[i, 0] = distance_matrix[i - 1, 0] + scoring_matrix_inplace(this[i - 1], '_', scoring_points)
    for j in range(1, len(that) + 1):
        distance_matrix[0, j] = distance_matrix[0, j - 1] + scoring_matrix_inplace('_', that[j - 1], scoring_points)

    for i in range(1, len(this) + 1):
        for j in range(1, len(that) + 1):
            distance_matrix[i, j] = max(
                distance_matrix[i - 1, j] + scoring_matrix_inplace(this[i - 1], '_', scoring_points),
                distance_matrix[i, j - 1] + scoring_matrix_inplace('_', that[j - 1], scoring_points),
                distance_matrix[i - 1, j - 1] + scoring_matrix_inplace(this[i - 1], that[j - 1], scoring_points)
            )

    # function returns table and global alignment score
    # alignment score is in cell (n,m) of the matrix
    return distance_matrix, distance_matrix[len(this), len(that)]


def traceback(this, that, distance_matrix, scoring_points):
    # initializing starting position cell(n,m)
    i = len(this)
    j = len(that)

    # initializing strings we use to represent alignments in x, y, edit transcript and global alignment
    ax, ay, am, tr = '', '', '', ''

    # exit condition is when we reach cell (0,0)
    while i > 0 or j > 0:

        # calculating diagonal, horizontal and vertical scores for current cell
        d, v, h = -100, -100, -100

        if i > 0 and j > 0:
            delta = 1 if this[i - 1] == that[j - 1] else 0
            d = distance_matrix[i - 1, j - 1] + scoring_matrix_inplace(this[i - 1], that[j - 1], scoring_points)  # diagonal movement
        if i > 0:
            v = distance_matrix[i - 1, j] + scoring_matrix_inplace(this[i - 1], '_', scoring_points)  # vertical movement
        if j > 0:
            h = distance_matrix[i, j - 1] + scoring_matrix_inplace('_', that[j - 1], scoring_points)  # horizontal movement

        # backtracking to next (previous) cell
        if d >= v and d >= h:
            ax += this[i - 1]
            ay += that[j - 1]
            if delta == 1:
                tr += 'M'
                am += '|'
            else:
                tr += 'R'
                am += ' '
            i -= 1
            j -= 1
        elif v >= h:
            ax += this[i - 1]
            ay += '_'
            tr += 'D'
            am += ' '
            i -= 1
        else:
            ay += that[j - 1]
            ax += '_'
            tr += 'I'
            am += ' '
            j -= 1

    alignment = '\n'.join([ax[::-1], am[::-1], ay[::-1]])
    return alignment, tr[::-1]



from genomics.seed_extend import *
from Bio import SeqIO
import logging
import click
import time
import csv
import sys

logger = logging.getLogger(__name__)


def import_or_generate_scoring_points(match, mismatch, gap):
    return {
        'M': match,
        'Ti': mismatch,
        'Tv': mismatch,
        'G': gap
    }


def import_file(path, file_type):
    return list(map(lambda r: str(r.seq), SeqIO.parse(path, file_type)))


def import_fasta_fastq(fasta_path='../data/example_human_reference.fasta', fastq_path='../data/example_human_Illumina.pe_1.fastq'):
    return import_file(fasta_path, file_type='fasta'), import_file(fastq_path, file_type='fastq')


def write_results_to_csv_file(file_name, results):
    with open(file_name, mode='w') as result_file:
        writer = csv.writer(result_file)
        writer.writerow(['start', 'end', 'alignment-score', 'transcription'])
        for result in results:
            writer.writerow(list(result))


def enable_logging(level=logging.INFO):
    _logger = logging.getLogger()
    _logger.setLevel(level=level)
    handler = logging.StreamHandler(sys.stdout)
    handler.setLevel(level=level)
    _logger.addHandler(handler)


@click.group()
def cli():
    pass


@cli.command()
@click.option('--fasta_path', type=str, required=True)
@click.option('--fastq_path', type=str, required=True)
@click.option('--margin', type=int, default=2)
@click.option('--match', type=int, required=True)
@click.option('--mismatch', type=int, required=True)
@click.option('--gap', type=int, required=True)
@click.option('--seed_length', type=int, default=10)
def run(fasta_path, fastq_path, margin, match, mismatch, gap, seed_length):
    try:
        enable_logging()
        logger.info(f'Starting giseed with params fasta_path={fastq_path} '
                    f'fastq_path={fastq_path} margin={margin} match={match} '
                    f'mismatch={mismatch} gap={gap} seed_length={seed_length}')

        scoring_points = import_or_generate_scoring_points(match, mismatch, gap)

        logger.info('Importing fasta and fastq files')
        references, reads = import_fasta_fastq(fasta_path, fastq_path)
        logger.info('Finished importing fasta and fastq files')

        logger.info('Starting seed and extend')
        results = seed_and_extend(scoring_points, references, reads, margin, seed_length)
        logger.info('Finished seed and extend')

        logger.info('Writing results to csv file')
        write_results_to_csv_file(f'results-{time.time()}.csv', results)
        logger.info('Finished writing results to csv file')
    except Exception as exc:
        logger.error('Following exception occurred ', exc)


if __name__ == '__main__':
    cli()



from genomics.main import *
import pytest


# BWT & FM Index
@pytest.mark.parametrize("example, result", [
    ('', []),
    ('$', ['$']),
    ('banana$', ['banana$', 'anana$b', 'nana$ba', 'ana$ban', 'na$bana', 'a$banan', '$banana'])
])
def test_rotations(example, result):
    assert rotations(example) == result


@pytest.mark.parametrize("example, result", [
    ('', []),
    ('$', ['$']),
    ('banana$', ['$banana', 'a$banan', 'ana$ban', 'anana$b', 'banana$', 'na$bana', 'nana$ba'])
])
def test_bwm(example, result):
    assert bwm(example) == result


def test_make_occurrences_matrix():
    test_occ_matrix, test_tots = make_occurrences_matrix(bwt_via_bwm('banana$'))
    assert test_occ_matrix == {
        'a': [1, 1, 1, 1, 1, 2, 3],
        'n': [0, 1, 2, 2, 2, 2, 2],
        'b': [0, 0, 0, 1, 1, 1, 1],
        '$': [0, 0, 0, 0, 1, 1, 1]
    }
    assert test_tots == {'a': 3, 'n': 2, 'b': 1, '$': 1}


@pytest.mark.parametrize("example, result", [
    ({'a': 3, 'n': 2, 'b': 1, '$': 1}, {'$': (0, 0), 'a': (1, 3), 'b': (4, 4), 'n': (5, 6)})
])
def test_first_col(example, result):
    assert first_col(example) == result


@pytest.mark.parametrize("example_word, example_query, result_start, result_end", [
    ('banana$', 'ana', 2, 3),
    ('bananaban$', 'ban', 5, 6)
])
def test_calculate_start_end_range(example_word, example_query, result_start, result_end):
    test_occ_matrix, test_tots = make_occurrences_matrix(bwt_via_bwm(example_word))
    test_c = first_col(test_tots)
    test_start, test_end = calculate_start_end_range(test_c, test_occ_matrix, example_query)
    assert (test_start, test_end) == (result_start, result_end)

# Global Alignment


scoring_points = {'M': 1, 'Ti': -1, 'Tv': -3, 'G': -7}
x = 'TACGTCAGC'
y = 'TATGTCATGC'
expected_distance = numpy.array([[0, -7, -14, -21, -28, -35, -42, -49, -56, -63, -70],
                                 [-7, 1, -6, -13, -20, -27, -34, -41, -48, -55, -62],
                                 [-14, -6, 2, -5, -12, -19, -26, -33, -40, -47, -54],
                                 [-21, -13, -5, 1, -6, -13, -18, -25, -32, -39, -46],
                                 [-28, -20, -12, -6, 2, -5, -12, -19, -26, -31, -38],
                                 [-35, -27, -19, -11, -5, 3, -4, -11, -18, -25, -32],
                                 [-42, -34, -26, -18, -12, -4, 4, -3, -10, -17, -24],
                                 [-49, -41, -33, -25, -19, -11, -3, 5, -2, -9, -16],
                                 [-56, -48, -40, -32, -24, -18, -10,  -2,   2,  -1,  -8],
                                 [-63, -55, -47, -39, -31, -25, -17,  -9,  -3,  -1,   0]])


@pytest.mark.parametrize("example_this, example_that, result", [
    ('C', '_', -7),
    ('C', 'G', -3),
    ('C', 'A', -3),
    ('C', 'T', -1),
    ('C', 'C', 1),

])
def test_scoring_matrix_inplace(example_this, example_that, result):
    assert scoring_matrix_inplace(example_this, example_that, scoring_points) == result


def test_global_alignment():
    expected_alignment_score = 0
    distances, alignment_score = global_alignment(x, y, scoring_points)
    assert (expected_distance == distances).all()
    assert expected_alignment_score == alignment_score


def test_traceback():
    alignment, transcript = traceback(x, y, expected_distance, scoring_points)
    expected_alignment =\
        '''TACGTCA_GC
|| |||| ||
TATGTCATGC'''  # really unfortunate formatting, consider removing since i think that this check is pointless
    assert "MMRMMMMIMM" == transcript
    assert expected_alignment == alignment


def test_fasta_fastq_import():
    test_references, test_reads = import_fasta_fastq()
    assert 1 == len(test_references)
    assert 172 == len(test_reads)


@pytest.mark.parametrize("example_this, example_that, result", [
    ('ACTGTAAC', 'GTTACAGT', True),
    ('AA', 'TT', True),
    ('', '', True),
    ('', 'A', False),
    ('A', 'C', False),
    ('A', 'G', False),
    ('ACTGTAAC', 'TGACATTG', False),

])
def test_reverse_and_complement(example_this, example_that, result):
    assert (reverse_and_complement(example_this) == example_that) == result



import logging

logger = logging.getLogger(__name__)


def rotations(t):
    """ Return list of rotations of input string t """
    words = list(t)
    all_rotations = []

    for i in range(len(words)):

        logger.info(f'Processing iteration {i} in rotations')

        word = t[-1] + t[:-1]
        new = ''.join(word)
        t = new
        all_rotations.append(''.join(word))
        i += 1
    return all_rotations



def bwm(t):
    """ Return lexicographically sorted list of t’s rotations """
    return sorted(rotations(t))


def bwt_via_bwm(t):
    """ Given T, returns BWT(T) by way of the BWM """
    return ''.join(map(lambda x: x[-1], bwm(t)))


def make_occurrences_matrix(bw):
    """ Given BWT string bw, returns a map of lists. Keys are
    characters and lists are cumulative # of occurrences up to and
    including the row. """
    tots = {}
    occ_matrix = {}
    for c in bw:
        if c not in tots:
            tots[c] = 0
            occ_matrix[c] = []
    for c in bw:
        tots[c] += 1
        for c in tots.keys():
            occ_matrix[c].append(tots[c])
    return occ_matrix, tots


def first_col(tots):
    """ Return map from character to the range of rows prefixed by the
    character. """
    first = {}
    totc = 0
    for c, count in sorted(tots.items()):
        first[c] = (totc, totc + count - 1)
        totc += count
    return first


def make_suffix_array(s):
    """ Given T return suffix array SA(T).  We use Python's sorted
        function here for simplicity, but we can do better. """
    satups = sorted([(s[i:], i) for i in range(len(s))])
    # Extract and return just the offsets
    return list(map(lambda x: x[1], satups))


def bwt_via_sa(t):
    """ Given T, returns BWT(T) by way of the suffix array. """
    bw = []
    for si in make_suffix_array(t):
        if si == 0:
            bw.append('$')
        else:
            bw.append(t[si - 1])
    return ''.join(bw)


def calculate_start_end_range(c, occ, query):
    try:
        reversed_query = query[::-1]
        start, end = c[reversed_query[0]]
        for ix in range(1, len(reversed_query)):
            if start > end:
                return -1, -1
            start = c[reversed_query[ix]][0] + occ[reversed_query[ix]][start - 1]
            end = c[reversed_query[ix]][0] + occ[reversed_query[ix]][end] - 1
        return start, end
    except Exception as e:
        return -1, -1


def find_all_query_positions_in_word_via_suffix_arr(start, end, suff_arr):
    query_positions_in_word = []
    while start <= end:
        query_positions_in_word.append(suff_arr[start])
        start += 1
    return sorted(query_positions_in_word)



from distutils.core import setup
import setuptools
setup(
    name='giseed',
    packages=['genomics'],
    py_modules=['genomics.main'],
    version='1.0.5',
    license='MIT',
    description='Tool developed for GI course on ETF Belgrade',
    author='Marko Aleksandar Marko',
    author_email='',
    url='https://github.com/accchi/GI-project',
    keywords=['bwt', 'genomics', 'seed&extend'],
    install_requires=[
      'biopython==1.78',
      'numpy==1.20.3',
      'pytest==6.2.4',
      'click==8.0.0'
    ],
    entry_points={
      'console_scripts': [
        'giseed=genomics.main:cli'
      ]
    }
)
